import multiprocessing
import sys
import matplotlib
# Do not use X environment otherwise graph don't work
matplotlib.use('Agg')
import networkx as nx
from matplotlib import pyplot as plt
from subprocess import check_output

global validData

def main():
    global validData

    # Number of processes
    numProcesses = 6
    # Minimum number of peers needed
    numPeers = 3
    # A socket has to have sent at least this number of packets.
    numMinPackets = 1
    # Pcap to be loaded
    filePcap = sys.argv[1]
    # Create test set (this will skip the step 1 checks)
    validData = int(sys.argv[2])

    print("-- Parsing --")
    print(("Parsing {} with {} processes, minimum packets is {} and minimum" +
          " peers is {}").format(filePcap,
                                numProcesses,
                                numMinPackets,
                                numPeers))

    # Get base filename
    fileNameBase = filePcap.split("/")[-1].split(".")[0]

    # Run tshark
    print("Running pcap through tshark")
    tshark = ["tshark",
              "-r",
              filePcap,
              "-Y",
              "ip.version == 4 && udp",
              "-T",
              "fields",
              "-e",
              "frame.number",
              "-e",
              "frame.len",
              "-e",
              "ip.src",
              "-e",
              "ip.dst",
              "-e",
              "ip.len",
              "-e",
              "udp.srcport",
              "-e",
              "udp.dstport",
              "-e",
              "udp.length",
              "-E",
              "separator=,"]
    packetsTshark = check_output(tshark).decode("ASCII").split("\n")

    # Parsing
    print("Parsing packets")
    packets = []
    for packet in packetsTshark:
        if packet == "":
            continue
        packets.append(packet.split(","))

    # Create multiprocessing pool
    p = multiprocessing.Pool(int(numProcesses))

    # Start multiple processed parsing for maximum speed
    results = p.map(parsePackets, chunkIt(packets, numProcesses))

    # Merge results
    outputs = [result for result in results]
    packetsSuspicious = []
    socketsNotSuspicious = []
    socketsSuspicious = {}
    for output in outputs:
        packetsSuspicious.extend(output[0])
        socketsNotSuspicious.extend(output[1])
        for key in output[2]:
            socketsSuspicious[key] = 1

    # Clean socket lists
    print("Cleaning socket lists.")
    socketsNotSuspicious, socketsSuspicious = \
        cleanSockets(socketsNotSuspicious,
                     socketsSuspicious)

    # Count frequencies and remove sockets that do not have numMinPackets
    print("Counting frequencies.")
    socketPacketLengthFrequency, packetsSuspicious, socketsNotSuspicious, \
        socketsSuspicious = getFrequencies(packetsSuspicious,
                                           socketsSuspicious,
                                           socketsNotSuspicious,
                                           numMinPackets)

    # Count peers and remove if less than numPeers
    print("Removing sockets with too few peers.")
    socketPacketLengthFrequency, socketsNotSuspicious, socketsSuspicious = \
        countPeers(socketPacketLengthFrequency,
                   socketsNotSuspicious,
                   socketsSuspicious,
                   numPeers)

    # Print results
    # FIXME packetsSuspicious print hier niet een goede waarde omdat hij niet
    # geupdate is na de vorige stap
    nodes, edges = printResults(socketPacketLengthFrequency,
                                packetsSuspicious,
                                socketsNotSuspicious,
                                socketsSuspicious,
                                packets)

    print("-- Files created --")
    # Create graph image
    graphFileName = "img/" + fileNameBase + "_graph.svg"
    createGraph(nodes, edges, graphFileName)

    # Output data
    if(len(socketPacketLengthFrequency) > 0):
        fileNameOutput = "data/" + fileNameBase + "_bof.csv"
        print("Output data to", fileNameOutput)
        outputBagOfFeatures(socketPacketLengthFrequency, fileNameOutput)
    else:
        print("No file outputed")


# Source:
# https://stackoverflow.com/questions/2130016/splitting-a-list-of-into-n-parts-of-approximately-equal-length
def chunkIt(seq, num):
    avg = len(seq) / float(num)
    out = []
    last = 0.0

    while last < len(seq):
        out.append(seq[int(last):int(last + avg)])
        last += avg

    return out


def cleanSockets(socketsNotSuspicious, socketsSuspicious):
    socketsNotSuspicious = list(set(socketsNotSuspicious))

    # Not suspicious has precedence over suspicious: as a single non suspicious
    # packet can declare the socket itself to be non suspicious.
    # Clean up this mess.
    toBeCleaned = []
    for socket in socketsSuspicious:
        if socket in socketsNotSuspicious:
            toBeCleaned.append(socket)
    for socket in toBeCleaned:
        del socketsSuspicious[socket]

    return [socketsNotSuspicious, socketsSuspicious]


def parsePackets(packets):
    global validData

    packetsSuspicious = []
    socketsNotSuspicious = []
    socketsSuspicious = {}

    for packet in packets:
        # Get Ethernet info
        # packetNum = int(packet[0])
        packetLen = int(packet[1])

        # Get IP info
        ipSrc = packet[2]
        ipDst = packet[3]
        ipLen = int(packet[4])

        # Get UDP info
        portSrc = int(packet[5])
        portDst = int(packet[6])
        udpLen = int(packet[7])

        # Socket
        socket = (ipSrc, ipDst, portSrc, portDst)

        # Check if this packet could be Freenet traffic
        notSuspicious = False

        # Step #1
        # 1. Port numbers have to be higher than 1024
        if portSrc <= 1024 or portDst <= 1024:
            notSuspicious = True

        # 2. Maximum packet size
        if packetLen > 1280:
            notSuspicious = True

        # 3. Minimum packet size
        if packetLen < 92:
            notSuspicious = True

        # 4. Maximum UDP payload size plus UDP header
        if udpLen > 1240:
            notSuspicious = True

        # 5. Minimum UDP payload size plus UDP header
        if udpLen < 72:
            notSuspicious = True

        # This can be used to generate test data
        #if validData == -1 and notSuspicious:
        #    continue

        # If not suspicious or the socket was already known not to be
        # suspicious.
        if notSuspicious or socket in socketsNotSuspicious:
            # Add to notSuspicious list if not already in there
            if socket not in socketsNotSuspicious:
                socketsNotSuspicious.append(socket)
            continue

        # Append suspicious packet to list
        packetsSuspicious.append([socket, packetLen, ipLen,  udpLen])

        # Add socket to suspicious dict
        if socket in socketsSuspicious:
            continue
        else:
            socketsSuspicious[socket] = 1

    return [packetsSuspicious, socketsNotSuspicious, socketsSuspicious]


def countPeers(socketPacketLengthFrequency,
               socketsNotSuspicious,
               socketsSuspicious,
               numPeers):
    # Step #2
    # Count the number of portDst per unique portSrc
    counted = {}
    for ipSrc, ipDst, portSrc, portDst in socketsSuspicious:
        if ipSrc in counted:
            if portSrc in counted[ipSrc]:
                counted[ipSrc][portSrc].append(portDst)
            else:
                counted[ipSrc][portSrc] = [portDst]
        else:
            counted[ipSrc] = {}
            counted[ipSrc][portSrc] = [portDst]

    # Find ipSrc which have numPeers or less portDst per portSrc
    foundStep2 = []
    for ipAddr, portsSrc in counted.items():
        # Loop over source ports
        for portSrc, portDsts in portsSrc.items():
            if len(portDsts) < numPeers:
                # Find matching socket
                for socket in socketsSuspicious:
                    socketIpSrc = socket[0]
                    socketIpPortSrc = socket[2]
                    if socketIpSrc == ipAddr and socketIpPortSrc == portSrc:
                        foundStep2.append(socket)

    # Remove the found sockets as there are not suspicious enough
    for socket in foundStep2:
        # Delete
        del socketsSuspicious[socket]
        del socketPacketLengthFrequency[socket]
        # Add to not suspicious sockets
        socketsNotSuspicious.append(socket)

    return [socketPacketLengthFrequency,
            socketsNotSuspicious,
            socketsSuspicious]


# Step #3
def getFrequencies(packetsSuspicious,
                   socketsSuspicious,
                   socketsNotSuspicious,
                   numMinPackets):
    # get frequency of packet lengths per suspicious socket
    socketPacketLengthFrequency = {}
    for packet in packetsSuspicious:
        for socket in socketsSuspicious:
            if packet[0] == socket:
                # Add socket to socketPacketLengthFrequency dict
                if socket not in socketPacketLengthFrequency:
                    socketPacketLengthFrequency[socket] = {}
                    socketPacketLengthFrequency[socket]["packetLen"] = {}
                    socketPacketLengthFrequency[socket]["ipLen"] = {}
                    socketPacketLengthFrequency[socket]["udpLen"] = {}
                # Add lengths to socket
                if packet[1] in socketPacketLengthFrequency[socket]["packetLen"]:
                    socketPacketLengthFrequency[socket]["packetLen"][packet[1]] += 1
                else:
                    socketPacketLengthFrequency[socket]["packetLen"][packet[1]] = 1
                if packet[2] in socketPacketLengthFrequency[socket]["ipLen"]:
                    socketPacketLengthFrequency[socket]["ipLen"][packet[2]] += 1
                else:
                    socketPacketLengthFrequency[socket]["ipLen"][packet[2]] = 1
                if packet[3] in socketPacketLengthFrequency[socket]["udpLen"]:
                    socketPacketLengthFrequency[socket]["udpLen"][packet[3]] += 1
                else:
                    socketPacketLengthFrequency[socket]["udpLen"][packet[3]] = 1

    # A socket should not be marked suspicious if it has sent only numMinPackets packets.
    # And a socket should also have received at least one packet from the other side.
    # This means that there should be a socket in the list which corresponds
    # with the other side of the connection, the counterSocket.
    remove = []
    for socket, value in socketPacketLengthFrequency.items():
        counterSocket = (socket[1], socket[0], socket[3], socket[2])

        # Check if counterSocket is in the list also
        if counterSocket not in socketPacketLengthFrequency.keys():
            remove.append(socket)
            del socketsSuspicious[socket]

        # Remove if not enough packets
        if len(value["packetLen"]) <= numMinPackets:
            if socket not in remove:
                remove.append(socket)
                del socketsSuspicious[socket]

    # Also remove from socket packet length frequency dictionary
    for socket in remove:
        del socketPacketLengthFrequency[socket]

    return [socketPacketLengthFrequency,
            packetsSuspicious,
            socketsNotSuspicious,
            socketsSuspicious]


def outputBagOfFeatures(socketPacketLengthFrequency, fileNameOutput):
    global validData

    # lengthTypes = ["packetLen", "udpLen", "ipLen"]
    lengthTypes = ["ipLen"]
    for socket, distribution in socketPacketLengthFrequency.items():
        # Loop over length types
        for lengthType in lengthTypes:
            # Add missing values
            for length in range(92, 1281):
                if length not in distribution[lengthType]:
                    distribution[lengthType][length] = 0

    # Write to file
    w = open(fileNameOutput, "w")
    for key, val in socketPacketLengthFrequency.items():
        for lengthType in lengthTypes:
            line = key[0] + ","
            line += key[1] + ","
            line += str(key[2]) + ","
            line += str(key[3]) + ","
            line += lengthType + ","
            total = 0
            # Count total amount of packets
            for length, frequency in val[lengthType].items():
                total += frequency
            # Write percentage per length to file
            for length, frequency in val[lengthType].items():
                line += str(frequency/total) + ","
            # Last column is 1 if it is valid Freenet traffic
            line += str(validData)
            # Write line to file
            w.write(line + "\n")


def printResults(socketPacketLengthFrequency, packetsSuspicious, socketsNotSuspicious, socketsSuspicious, packets):
    print("")
    print("-- Results --")

    # Print statistics
    print("Total packets", len(packets))
    print("Not suspicious packets", len(packets) - len(packetsSuspicious))
    print("Suspicious packets", len(packetsSuspicious))
    print("")

    print("Total sockets", len(socketsSuspicious) + len(socketsNotSuspicious))
    print("Not suspicious sockets", len(socketsNotSuspicious))
    print("Suspicious sockets", len(socketsSuspicious))
    print("")

    # Print suspicious sockets
    edges = []
    nodes = []
    for socket in sorted(socketsSuspicious):
        edges.append(socket[:2])
        nodes.append(socket[0])
        print(*socket)
    nodes = sorted(list(set(nodes)))

    return [nodes, edges]


# Step #4 remainder suspicious sockets can now be mapped in a graph
# Create graph if there are nodes found
def createGraph(nodes, edges, fileName):
    if len(edges) > 0:
        print("Creating graph", fileName)
        # Create graph
        G = nx.Graph(edges)

        # Use spring layout
        pos = nx.spring_layout(G, k=50)

        # Draw nodes
        nx.draw_networkx_nodes(G, pos, nodelist=nodes)

        # Draw edges
        nx.draw_networkx_edges(G, pos, edgelist=edges)

        # Draw labels
        labels = {}
        for node in nodes:
            labels[node] = node
        nx.draw_networkx_labels(G, pos, labels)

        # Increase margin
        plot_margin = 0.25
        x0, x1, y0, y1 = plt.axis()
        plt.axis((x0 - plot_margin/2,
                  x1 + plot_margin,
                  y0 - plot_margin,
                  y1 + plot_margin/2))

        # Remove axis
        plt.axis('off')

        # Save to file
        plt.savefig(fileName, format="SVG", bbox_inches="tight", frameOn=False)
    else:
        print("No graph created")


if "__main__":
    main()
